/* X509 Profile Exts management for libpki */

#include <libpki/pki.h>


/*?
 * \brief Adds extensions to a certificate according to the profile
 */

int PKI_X509_EXTENSIONS_cert_add_profile ( PKI_X509_PROFILE *conf, 
				PKI_CONFIG *oids, PKI_X509_CERT *x, PKI_TOKEN *tk ) {

	PKI_X509_EXTENSION *ext = NULL;

	int i = -1;
	int ext_num = -1;

	if ( !conf || !x || !x->value ) return PKI_ERR;

	ext_num = PKI_X509_PROFILE_get_exts_num ( conf );

	for ( i = 0; i < ext_num; i++ ) {
		if (( ext = PKI_X509_PROFILE_get_ext_by_num ( conf, i, tk )) != NULL ) {
			PKI_X509_CERT_add_extension ( x, ext );
		} else {
			PKI_log_debug ("Can not create EXTENSION number %d", i);
		}
	}

	return PKI_OK;
/*
	PKI_X509_EXTENSION *ext = NULL;

	// PKI_OID *oid = NULL;
	PKI_CONFIG_ELEMENT *curr = NULL;

	// PKI_CONFIG_ELEMENT *root_element = NULL;
	// PKI_CONFIG_ELEMENT *opt = NULL;

	xmlXPathContext *xpathCtx = NULL; 
	xmlXPathObject *xpathObj = NULL;
	xmlNodeSet *nodes = NULL;

	xmlChar oidSearchBuff[BUFF_MAX_SIZE];

	int size = 0;
	int i = 0;

	if( !conf || !x ) return (PKI_ERR);

	snprintf((char *) oidSearchBuff, BUFF_MAX_SIZE, 
		(const char *) "/%s:profile/%s:extensions/%s:extension",
				PKI_NAMESPACE_PREFIX, 
				PKI_NAMESPACE_PREFIX, 
				PKI_NAMESPACE_PREFIX);

	xpathCtx = xmlXPathNewContext(conf);
	if(xpathCtx == NULL) {
        	PKI_log_debug("ERROR, unable to create new XPath context!\n");
		return(PKI_ERR);
	}


	xmlXPathRegisterNs(xpathCtx, (xmlChar *) PKI_NAMESPACE_PREFIX, 
				(xmlChar *) PKI_NAMESPACE_HREF);

	xpathObj = xmlXPathEvalExpression( oidSearchBuff, xpathCtx);
	if( xpathObj == NULL ) {
		xmlXPathFreeContext(xpathCtx);
		return(PKI_ERR);
	}

	nodes = xpathObj->nodesetval;
	if( nodes ) {
		size = nodes->nodeNr;
	}

	for( i = 0; i < size; i++ ) {
		if( nodes->nodeTab[i]->type == XML_ELEMENT_NODE ) {

			curr = nodes->nodeTab[i];
			ext = PKI_X509_EXTENSION_value_new_profile ( conf, 
								oids, curr );
			if( ext ) {
				int r;

				r = PKI_X509_CERT_add_extension( x, ext );
				if( r == PKI_ERR ) {
					PKI_log_debug("ERROR, can not add extension!\n"); 
				} else {
					PKI_log_debug("INFO, ext added ok!\n");
				}
			} else {
				PKI_log_debug("ERROR, extension is NULL!!!\n");
			}
		}
	}

	xmlXPathFreeObject(xpathObj);
	xmlXPathFreeContext(xpathCtx);

	return (PKI_OK);
	*/
}


int PKI_X509_EXTENSIONS_req_add_profile ( PKI_X509_PROFILE *conf, 
				PKI_CONFIG *oids, PKI_X509_REQ *req, PKI_TOKEN *tk ) {

	PKI_X509_EXTENSION *ext = NULL;
	int i = -1;
	int ret = -1;
	int ext_num = -1;

	if ( !conf || !req || !req->value ) return PKI_ERR;

	ext_num = PKI_X509_PROFILE_get_exts_num ( conf );

	PKI_log_debug(">>>>>>>>>>>>>>> EXT NUM: %d <<<<<<<<<<<<<<<", ext_num );

	for ( i = 0; i < ext_num; i++ ) {
		if (( ext = PKI_X509_PROFILE_get_ext_by_num ( conf, i, tk )) != NULL ) {
			ret = PKI_X509_REQ_add_extension ( req, ext );
			PKI_log_debug("Extension %d added, result is %d", i, ret );
		} else {
			PKI_log_debug ("Can not create EXTENSION number %d", i);
		}
	}

	return PKI_OK;
/*
	PKI_CONFIG_ELEMENT *curr = NULL;


	xmlXPathContext *xpathCtx = NULL; 
	xmlXPathObject *xpathObj = NULL;
	xmlNodeSet *nodes = NULL;

	xmlChar oidSearchBuff[BUFF_MAX_SIZE];

	PKI_X509_EXTENSION_STACK *sk = NULL;

	int size = 0;
	int i = 0;

	if( !conf || !req || !req->value ) return (PKI_ERR);

	snprintf( (char *) oidSearchBuff, BUFF_MAX_SIZE, 
		"/%s:profile/%s:extensions/%s:extension",
			PKI_NAMESPACE_PREFIX, PKI_NAMESPACE_PREFIX, 
				PKI_NAMESPACE_PREFIX);

	xpathCtx = xmlXPathNewContext(conf);
	if(xpathCtx == NULL) {
        	PKI_log_debug("ERROR, unable to create new XPath context!\n");
		return(PKI_ERR);
	}


	xmlXPathRegisterNs(xpathCtx, (xmlChar *) PKI_NAMESPACE_PREFIX, 
				(xmlChar *) PKI_NAMESPACE_HREF);

	xpathObj = xmlXPathEvalExpression( oidSearchBuff, xpathCtx);
	if( xpathObj == NULL ) {
		xmlXPathFreeContext(xpathCtx);
		return(PKI_ERR);
	}

	nodes = xpathObj->nodesetval;
	if( nodes ) {
		size = nodes->nodeNr;
	}


	// sk = sk_X509_EXTENSION_new_null();
	if((sk = PKI_STACK_X509_EXTENSION_new()) == NULL ) {
		PKI_log_debug("ERROR, can not create a new stack of exts!\n");
	}

	for( i = 0; i < size; i++ ) {
		if( nodes->nodeTab[i]->type == XML_ELEMENT_NODE ) {
			// xmlChar *name = NULL;
			// xmlChar *val = NULL;

			curr = nodes->nodeTab[i];
			ext = PKI_X509_EXTENSION_value_new_profile ( conf, 
								oids, curr );
			if( ext ) {
				int r;

				r = PKI_STACK_X509_EXTENSION_push( sk, ext );
				if( r == PKI_ERR ) {
					PKI_log_debug("ERROR, can not add extension!\n"); 
				} else {
					PKI_log_debug("DEBUG, ext added ok!\n");
				}
			} else {
				PKI_log_debug("ERROR, ext is NULL!!!\n");
			}
		}
	}

	xmlXPathFreeObject(xpathObj);
	xmlXPathFreeContext(xpathCtx);

	if((PKI_X509_REQ_add_extension_stack( req->value, sk )) == PKI_ERR ) {
		PKI_log_debug("ERROR, can not add request extensions!\n");
		return (PKI_ERR);
	}

	return (PKI_OK);
*/
}


/*?
 * \brief Adds extensions to a CRL according to the profile passed as argument
 */

int PKI_X509_EXTENSIONS_crl_add_profile ( PKI_X509_PROFILE *conf, 
				PKI_CONFIG *oids, PKI_X509_CRL *crl, PKI_TOKEN *tk ) {

	PKI_X509_EXTENSION *ext = NULL;

	int i = -1;
	int ext_num = -1;

	if ( !conf || !crl || !crl->value ) return PKI_ERR;

	ext_num = PKI_X509_PROFILE_get_exts_num ( conf );

	for ( i = 0; i < ext_num; i++ ) {
		if (( ext = PKI_X509_PROFILE_get_ext_by_num ( conf, i, tk )) != NULL ) {
			PKI_X509_CRL_add_extension ( crl, ext );
		} else {
			PKI_log_debug ("Can not create EXTENSION number %d", i);
		}
	}

	return PKI_OK;

/*
	PKI_X509_EXTENSION *ext = NULL;

	// PKI_OID *oid = NULL;
	PKI_CONFIG_ELEMENT *curr = NULL;

	// PKI_CONFIG_ELEMENT *root_element = NULL;
	// PKI_CONFIG_ELEMENT *opt = NULL;

	xmlXPathContext *xpathCtx = NULL; 
	xmlXPathObject *xpathObj = NULL;
	xmlNodeSet *nodes = NULL;

	xmlChar oidSearchBuff[BUFF_MAX_SIZE];

	int size = 0;
	int i = 0;

	if( !conf || !crl ) return (PKI_ERR);

	snprintf((char *) oidSearchBuff, BUFF_MAX_SIZE, 
		(const char *) "/%s:profile/%s:extensions/%s:extension",
				PKI_NAMESPACE_PREFIX, 
				PKI_NAMESPACE_PREFIX, 
				PKI_NAMESPACE_PREFIX);

	xpathCtx = xmlXPathNewContext(conf);
	if(xpathCtx == NULL) {
        	PKI_log_debug("ERROR, unable to create new XPath context!\n");
		return(PKI_ERR);
	}


	xmlXPathRegisterNs(xpathCtx, (xmlChar *) PKI_NAMESPACE_PREFIX, 
				(xmlChar *) PKI_NAMESPACE_HREF);

	xpathObj = xmlXPathEvalExpression( oidSearchBuff, xpathCtx);
	if( xpathObj == NULL ) {
		xmlXPathFreeContext(xpathCtx);
		return(PKI_ERR);
	}

	nodes = xpathObj->nodesetval;
	if( nodes ) {
		size = nodes->nodeNr;
	}

	for( i = 0; i < size; i++ ) {
		if( nodes->nodeTab[i]->type == XML_ELEMENT_NODE ) {

			curr = nodes->nodeTab[i];
			ext = PKI_X509_EXTENSION_value_new_profile ( conf, 
								oids, curr );
			if( ext ) {
				int r;

				r = PKI_X509_CRL_add_extension( crl, ext );
				if( r == PKI_ERR ) {
					PKI_log_debug("ERROR, can not add extension!\n"); 
				} else {
					PKI_log_debug("INFO, ext added ok!\n");
				}
			} else {
				PKI_log_debug("ERROR, extension is NULL!!!\n");
			}
		}
	}

	xmlXPathFreeObject(xpathObj);
	xmlXPathFreeContext(xpathCtx);

	return (PKI_OK);
	*/
}

