/* ERR String definitions
 * 
 * NOTE: When adding a new error, please check that the PKI_ERR_
 *       and the corresponding string description match the error!
 */

#ifndef _LIBPKI_ERR_H
#define _LIBPKI_ERR_H

typedef enum {
	/* General Errors */
	PKI_ERR_UNKNOWN 	= 0,
	PKI_ERR_GENERAL,
	PKI_ERR_MEMORY_ALLOC,
	PKI_ERR_POINTER_NULL,
	PKI_ERR_PARAM_NULL,
	PKI_ERR_PARAM_TYPE,
	/* PKI MEM Errors */
	PKI_ERR_MEM_,
	/* URI related Errors */
	PKI_ERR_URI_UNSUPPORTED,
	PKI_ERR_URI_GENERAL,
	PKI_ERR_URI_PARSE,
	PKI_ERR_URI_OPEN,
	PKI_ERR_URI_CLOSE,
	PKI_ERR_URI_READ,
	PKI_ERR_URI_WRITE,
	PKI_ERR_URI_DNS,
	PKI_ERR_URI_SSL,
	PKI_ERR_URI_SSL_TRUST,
	/* HSM Related */
	PKI_ERR_HSM_INIT,
	PKI_ERR_HSM_LOGIN,
	PKI_ERR_HSM_SET_ALGOR,
	PKI_ERR_HSM_SET_SLOT,
	PKI_ERR_HSM_KEYPAIR_LOAD,
	PKI_ERR_HSM_KEYPAIR_STORE,
	PKI_ERR_HSM_KEYPAIR_IMPORT,
	PKI_ERR_HSM_KEYPAIR_EXPORT,
	PKI_ERR_HSM_,
	/* Configuration Related */
	PKI_ERR_CONFIG_MISSING,
	PKI_ERR_CONFIG_LOAD,
	PKI_ERR_CONFIG_SAVE,
	PKI_ERR_CONFIG_,
	/* Profile Related */
	PKI_ERR_X509_PROFILE_,
	/* Token Related */
	PKI_ERR_TOKEN_INIT,
	PKI_ERR_TOKEN_KEYPAIR_LOAD,
	PKI_ERR_TOKEN_SET_CRED,
	PKI_ERR_TOKEN_USE_SLOT,
	PKI_ERR_TOKEN_PROFILE_LOAD,
	PKI_ERR_TOKEN_SET_ALGOR,
	PKI_ERR_TOKEN_,
	/* Key Operations */
	PKI_ERR_X509_KEYPAIR_,
	/* Certificate Operations */
	PKI_ERR_X509_CERT_,
	PKI_ERR_X509_CERT_VERIFY_,
	/* Request Operations */
	PKI_ERR_X509_REQ_,
	/* CRL Errors */
	PKI_ERR_X509_CRL,
	/* OCSP Ops */
	PKI_ERR_OCSP_,
	/* PRQP Ops */
	PKI_ERR_PRQP_,
	/* PKI Message Operations */
	PKI_ERR_MSG_,
	/* Enrollment Protocol Related */
	PKI_ERR_ENROLL_,
	/* Signatures Related Errors */
	PKI_ERR_SIGN_,
	PKI_ERR_SIGN_VERIFY,
} PKI_ERR_CODE;

typedef struct pki_err_st {
	PKI_ERR_CODE code;
	char *descr;
} PKI_ERR_ST;

/* We define the actual object of strings only if we are compiling
 * the pki_err.c file where __LIBPKI_ERR__ is defined */

#ifdef __LIBPKI_ERR__

const PKI_ERR_ST __libpki_errors_st[] = {
	/* General Errors */
	{ PKI_ERR_UNKNOWN, "Unknown Error" },
	{ PKI_ERR_GENERAL, "General Error" },
	{ PKI_ERR_MEMORY_ALLOC, "Memory Allocation Error" },
	{ PKI_ERR_POINTER_NULL, "Null Memory Pointer" },
	{ PKI_ERR_PARAM_NULL, "Null Parameter" },
	{ PKI_ERR_PARAM_TYPE, "Wrong Paramenter Type" },
	/* PKI MEM Errors */
	{ PKI_ERR_MEM_, "" },
	/* URI Related Operations */
	{ PKI_ERR_URI_UNSUPPORTED, "Unsupported URI Schema" },
	{ PKI_ERR_URI_GENERAL, "URI General Error" },
	{ PKI_ERR_URI_PARSE, "URI parsing Error" },
	{ PKI_ERR_URI_OPEN, "File Open Error" },
	{ PKI_ERR_URI_CLOSE, "File Close Error" },
	{ PKI_ERR_URI_READ, "File Read Error" },
	{ PKI_ERR_URI_WRITE, "File Write Error" },
	{ PKI_ERR_URI_DNS, "DNS Error" },
	{ PKI_ERR_URI_SSL, "SSL Connection Error" },
	{ PKI_ERR_URI_SSL_TRUST, "Untrusted SSL Connection Error" },
	/* HSM Related */
	{ PKI_ERR_HSM_INIT, "Can not init HSM" },
	{ PKI_ERR_HSM_LOGIN, "Can not login in HSM" },
	{ PKI_ERR_HSM_SET_ALGOR, "Error while chosing algorithm in HSM" },
	{ PKI_ERR_HSM_SET_SLOT, "Error while setting HSM slot to use" },
	{ PKI_ERR_HSM_KEYPAIR_LOAD, "Can not load HSM key" },
	{ PKI_ERR_HSM_KEYPAIR_STORE, "Can not store Key to HSM" },
	{ PKI_ERR_HSM_KEYPAIR_IMPORT, "Can not import Key to HSM" },
	{ PKI_ERR_HSM_KEYPAIR_EXPORT, "Can not export Key from HSM" },
	{ PKI_ERR_HSM_, "" },
	/* Configuration Related */
	{ PKI_ERR_CONFIG_MISSING, "Can not find Configuration file" },
	{ PKI_ERR_CONFIG_LOAD, "Error while loading configuration data" },
	{ PKI_ERR_CONFIG_SAVE, "Error while storing configuration data" },
	{ PKI_ERR_CONFIG_, "" },
	/* Profile Related */
	{ PKI_ERR_X509_PROFILE_, "" },
	/* Token Related */
	{ PKI_ERR_TOKEN_INIT, "Can not initialize Token" },
	{ PKI_ERR_TOKEN_SET_CRED, "Can not set Token credentials" },
	{ PKI_ERR_TOKEN_USE_SLOT, "Error while setting Token's Slot" },
	{ PKI_ERR_TOKEN_PROFILE_LOAD, "Can not load Token's Profile" },
	{ PKI_ERR_TOKEN_SET_ALGOR, "Error while setting Token's Algorithm" },
	{ PKI_ERR_TOKEN_KEYPAIR_LOAD,  "Can not load Token Key" },
	{ PKI_ERR_TOKEN_,  "" },
	/* Key Operations */
	{ PKI_ERR_X509_KEYPAIR_, "" },
	/* Certificate Operations */
	{ PKI_ERR_X509_CERT_, "" },
	{ PKI_ERR_X509_CERT_VERIFY_, "" },
	/* Request Operations */
	{ PKI_ERR_X509_REQ_, "" },
	/* CRL Errors */
	{ PKI_ERR_X509_CRL, "" },
	/* OCSP Ops */
	{ PKI_ERR_OCSP_, "" },
	/* PRQP Ops */
	{ PKI_ERR_PRQP_, "" },
	/* PKI Message Operations */
	{ PKI_ERR_MSG_, "" },
	/* Enrollment Protocol Related */
	{ PKI_ERR_ENROLL_, "" },
	/* Signatures Related Errors */
	{ PKI_ERR_SIGN_, "" },
	{ PKI_ERR_SIGN_VERIFY, "" },
};

static const int __libpki_err_size = sizeof ( __libpki_errors_st ) / sizeof ( PKI_ERR_ST );

#endif /* __LIBPKI_ERR__ */

int PKI_ERROR ( int err, char *info );

#endif
