/* openssl/pki_pkey.c */

#include <libpki/pki.h>

PKI_X509_KEYPAIR *PKI_X509_KEYPAIR_new_null () {
	return PKI_X509_new ( PKI_DATATYPE_X509_KEYPAIR, NULL );
}

void PKI_X509_KEYPAIR_free( PKI_X509_KEYPAIR *key ) {

	PKI_X509_free ( key );
	return;
/*
	HSM *hsm = NULL;

	if( !key || ! key->value ) return;

	hsm = key->hsm;

	if( hsm && hsm->callbacks && hsm->callbacks->keypair_free ) {
		return hsm->callbacks->keypair_free( key );
	}
		return HSM_OPENSSL_KEYPAIR_free ( key );
	}

	return (PKI_OK);
*/
}

void PKI_X509_KEYPAIR_free_void ( void *key ) {
	PKI_X509_free_void ( (PKI_X509_KEYPAIR *) key );
	return;
}

/*! \brief Generate a new Keypair with the passed label (required for
 *         PKCS#11 HSMs ) as target
 */

PKI_X509_KEYPAIR *PKI_X509_KEYPAIR_new( int type, int bits, char *label,
                                        PKI_CRED *cred, HSM *hsm ) {
	return HSM_X509_KEYPAIR_new ( type, bits, label, cred, hsm );
}

/*! \brief Generate a new Keypair with the passed URL (required for
 *         PKCS#11 HSMs ) as target
 */

PKI_X509_KEYPAIR *PKI_X509_KEYPAIR_new_url( int type, int bits, URL *url,
                                        PKI_CRED *cred, HSM *hsm ) {
	return HSM_X509_KEYPAIR_new_url ( type, bits, url, cred, hsm );
}

/*! \brief Returns a char * with a string representation of the Keypair
 */

char * PKI_X509_KEYPAIR_get_parsed ( PKI_X509_KEYPAIR *pkey ) {

	if( !pkey || !pkey->value ) return ( NULL );

	return ("To Be Implemented");
}

/*! \brief Returns the (unsigned char *) digest of a pubkey value */

PKI_DIGEST *PKI_X509_KEYPAIR_VALUE_pub_digest ( PKI_X509_KEYPAIR_VALUE *pkey,
							PKI_DIGEST_ALG *md ) {

	X509_PUBKEY *xpk = NULL;
	ASN1_BIT_STRING *key = NULL;
	PKI_DIGEST * ret = NULL;

	if (!pkey) return NULL;

	if( !md ) md = PKI_DIGEST_ALG_DEFAULT;

	if(!X509_PUBKEY_set(&xpk, pkey )) {
		PKI_log_debug("PKI_X509_KEYPAIR_pub_digest()::Error building X509 "
			"PUBKEY data");
		return NULL;
	}
	if((key = xpk->public_key ) == NULL ) {
		PKI_log_debug("PKI_X509_KEYPAIR_pub_digest()::No pubkey found!");
		return ( NULL );
	}

	if( key->length < 1 ) {
		PKI_log_debug("PKI_X509_KEYPAIR_pub_digest()::Pubkey len is 0!");
		return ( NULL );
	}

	if(( ret = PKI_DIGEST_new( md, key->data, 
					(size_t) key->length )) == NULL ) {
		PKI_log_debug("PKI_X509_KEYPAIR_pub_digest()::%s",
			ERR_error_string( ERR_get_error(), NULL ));
		return ( NULL );
	}

	return ( ret );
}

/*! \brief Returns the (unsigned char *) digest of the pubkey */

PKI_DIGEST *PKI_X509_KEYPAIR_pub_digest ( PKI_X509_KEYPAIR *k, 
							PKI_DIGEST_ALG *md) {

	if( !k || !k->value ) return ( NULL );

	return PKI_X509_KEYPAIR_VALUE_pub_digest ( k->value, md );

}

/*! \brief Returns the passed PKI_X509_KEYPAIR in PKCS#8 format */

PKI_MEM *PKI_X509_KEYPAIR_get_p8 ( PKI_X509_KEYPAIR *k ) {

	BIO *mem = NULL;
	PKI_MEM *ret = NULL;
	PKI_X509_KEYPAIR_VALUE *pkey = NULL;

	if ( !k || !k->value ) return NULL;

	pkey = k->value;

	if((mem = BIO_new(BIO_s_mem())) == NULL ) {
		return NULL;
	}

	if(i2d_PKCS8PrivateKeyInfo_bio(mem, (EVP_PKEY *) pkey) > 0 ) {
		if( BIO_flush ( mem ) <= 0 ) {
			PKI_log_debug("ERROR flushing mem");
		}
		ret = PKI_MEM_new_bio ( mem, NULL );
	}

	BIO_free ( mem );

	return ( ret );
}

/*! \brief Reads a PKI_X509_KEYPAIR from a PKCS#8 format */

PKI_X509_KEYPAIR *PKI_X509_KEYPAIR_new_p8 ( PKI_MEM *buf ) {

	PKI_log_debug ("To Be implemented!");
	return ( NULL );
}

