
#include <libpki/pki.h>

/* Global Variables - not a good coding practice, but for a small
   tool is not a problem (and it saves in complexity for signatures
   of the functions */

int verbose = 1;
int debug = 0;


/*------------------------------ Functions -----------------------------*/

char * prompt_pin( char * pt ) {
	char buf[1024];
	char *ret = NULL;
	int i;
	int k;
	char c;

	printf("%s", pt);
	memset(buf, 0, sizeof(buf));

	for(i = 0 ; i < sizeof(buf); i++ ) {
		// c = getch();
		k = scanf("%c", &c);
		buf[i] = c;
		if( c == '\n' || c == '\r' ) {
			buf[i] = '\x0';
			break;
		}
	}

	ret = (char *) PKI_Malloc ((size_t)(i+1));
	memcpy(ret, buf, (size_t) i+1);

	return ret;
}

char * prompt_str( char * pt ) {
	char buf[1024];
	char *ret = NULL;
	int i;
	int k;
	char c;

	printf("%s", pt);
	memset(buf, 0, sizeof(buf));

	for(i = 0 ; i < sizeof(buf); i++ ) {
		k = scanf("%c", &c);
		buf[i] = c;
		if( c == '\n' || c == '\r' ) {
			buf[i] = '\x0';
			break;
		}
	}

	ret = (char *) PKI_Malloc ((size_t) i+1);
	memcpy(ret, buf, (size_t) i+1);
	return ret;
}

void usage ( void ) {

	printf("\n\nPKI Token Tool - Massimiliano Pala <madwolf@openca.org>\n");
	printf("(c) 2006-2010 by Massimiliano Pala and OpenCA Labs\n");
	printf("OpenCA Licensed Software\n\n");

	printf("  USAGE: pki-tool cmd [options]\n\n");
	printf("  Where cmd is:\n");
	printf("  info            - Prints out information about the token\n");
	printf("  list            - List the names of available tokens\n");
	printf("  clear           - Deletes all the data on the token\n");
	printf("  delete          - Deletes objects (use -uri)\n");
	printf("  genkey          - Generates a new Keypair (def. RSA-SHA1)\n");
	printf("  genreq          - Generates a new X.509 PKCS#10 request\n");
	printf("  gencert         - Generates a new X.509 certificate\n");
	printf("  import          - Import an item in the token\n");
	printf("\n");

	printf("  Where Options are:\n");
	printf("  -token          - Name of the token to be used\n");
	printf("  -config <dir>   - Token config dir (HOME/.libpki/token.d)\n");
	printf("  -hsm <name>     - HSM name (HOME/.libpki/hsm.d)\n");
	printf("  -in <url>       - Input Data URI\n");
	printf("  -out <url>      - Output Data URI\n");
	printf("  -bits <num>     - Number of Bits\n");
	printf("  -type <objtype> - Type of Object\n");
	printf("  -algor <name>   - Algorithm to be used\n");
	printf("  -newkey         - Generate new keypair when using genreq\n");
	printf("  -uri <uri>      - URI of the item (key/cert/..) in the "
				    "token\n");
	printf("  -signkey <uri>  - URI of the cert-signing key\n");
	printf("  -signcert <uri> - URI of the cert-signing cert (CA)\n");
	printf("  -subject <dn>   - Distinguished Name (Subject)\n");
	printf("  -serial <num>   - Serial Number to use (gencert)\n");
	printf("  -profile <name> - Profile to use (gencert/genreq)\n");
	printf("  -profileuri <uri> - Profile URI to load (gencert/genreq)\n");
	printf("  -oidsuri <uri>  - OID files to load (gencert/genreq)\n");
	printf("  -days <num>     - Validity period (days)\n");
	printf("  -hours <num>    - Validity period (hours)\n");
	printf("  -mins <num>     - Validity period (mins)\n");
	printf("  -secs <num>     - Validity period (secs)\n");
	printf("  -selfsign       - Generate a self signed X.509 cert\n");
	printf("  -batch          - Batch mode (no prompt - assumes yes)\n");
	printf("  -verbose        - Writes additional info to stdout\n");
	printf("  -debug          - Enables Debugging info to stderr\n");

	printf("\n  Where Type of Object can be:\n");
	printf("   any            - Unknown type\n");
	printf("   key            - Keypair (Pub and Priv Keys)\n");
	printf("   pubkey         - Public Key\n");
	printf("   privkey        - Private Key\n");
	printf("   user           - User Certificates\n");
	printf("   ca             - CA Certificates\n");
	printf("   trusted        - Trusted Certificates (TA)\n");
	printf("   other          - Other Certificates\n");
	printf("   crl            - CRL\n");
	
	printf("\n");

	exit(1);
}

void version ( void ) {
	printf("PKI Tool::Version=%s\n", PACKAGE_VERSION );

	exit(0);
}

int gen_keypair ( PKI_TOKEN *tk, int bits, char *url_s, char *algor_opt,
								int batch ) {

	int algor_id = 0;
	char * prompt = NULL;

	URL *keyurl = NULL;

	if( bits == 0 ) {
		printf("\nERROR: needed option '-bits'\n\n");
		exit(1);
	}
	/*
	if( !url_s ) {
		printf("\nERROR: needed option '-keyurl <url>'\n\n");
		exit(1);
	}
	*/

	if ( url_s == NULL ) {
		if((url_s = tk->key_id) == NULL ) {
			url_s = "fd://1";
		}
	}
	keyurl = URL_new( url_s );

	if( keyurl == NULL ) {
		printf("\nERROR: can not parse URL [%s]\n\n", url_s );
		exit(1);
	}

	if( algor_opt == NULL ) {
		/* Set default algorithm to RSA-SHA1 */
		algor_id = PKI_ALGOR_RSA_SHA1;
	} else {
		PKI_ALGOR *algor = NULL;

		/* Get the Algor Id from the cmd line option */
		algor = PKI_ALGOR_get_by_name ( algor_opt );
		if( !algor ) {
			printf("\nERROR: Algorithm %s is not "
					"recognized!\n\n", algor_opt);
			exit(1);
		}
		algor_id = PKI_ALGOR_get_id ( algor );
	}

	if( verbose ) printf("\nSetting Token Algorithm to %s ... ", 
					PKI_ALGOR_ID_txt ( algor_id ) );

	if((PKI_TOKEN_set_algor ( tk, algor_id )) == PKI_ERR ) {
		printf("ERROR, can not set the crypto scheme!\n\n");
		exit(1);
	}
	if( verbose ) printf("Ok.\n");

	if( !batch ) {
		printf("\nThis will generate a new keypair on the "
							"Token.\n");
		printf("\nDetails:\n");
		printf("  - Algorithm ......: %s\n", 
					PKI_ALGOR_ID_txt( algor_id ));
		printf("  - Bits ...........: %d\n", bits );
		printf("  - Output .........: %s\n", keyurl->url_s );

		prompt = prompt_str ("\nAre you sure [y/N] ? ");
	}
	if( batch || (strncmp_nocase(prompt, "y", 1 ) == 0)) {
		
		int ret = PKI_OK;
		char *tmp_s = NULL;

		if(verbose) {
			printf("Generating new KeyPair ... ");
			fflush(stdout);
		}

		if((PKI_TOKEN_new_keypair_url( tk, bits, keyurl )) == PKI_ERR) {
			printf("Failed!\n");
			return (PKI_ERR);
		}
		if( verbose ) printf("Ok.\n");

		if( tk->type != HSM_TYPE_PKCS11 ) {
			if( keyurl ) {
				if( verbose ) {
					printf("Saving keypair to %s ... ",
						keyurl->url_s );
					fflush(stdout);
				}

				tmp_s = keyurl->url_s;
				ret = PKI_TOKEN_export_keypair( tk, tmp_s,
							PKI_DATA_FORMAT_PEM );
			} else {
				if((tmp_s = tk->key_id ) == NULL ) {
					tmp_s = "fd://1";
				}
				ret = PKI_TOKEN_export_keypair( tk, tk->key_id,
							PKI_DATA_FORMAT_PEM );
			}
			if( verbose ) printf("(%s) ... ", tmp_s );

			if ( ret == PKI_ERR ) {
				printf("ERROR, can not store key (%s)!\n\n", 
								tmp_s);
				return ( PKI_ERR );
			}
			if( verbose ) printf("Ok!\n");
		}
	} else {
		printf("\nOperation Cancelled.\n\n");
	}

	return ( PKI_OK );
}

int main (int argc, char *argv[] ) {

	PKI_TOKEN *tk = NULL;
	PKI_X509_PROFILE *prof =  NULL;

	unsigned long i = 0;
	unsigned long start = 2;
	char * prompt = NULL;

	char * cmd = NULL;
	char * token_name = NULL;
	char * infile = NULL;
	char * outfile = NULL;
	char * config = NULL;

	int log_level = 0;
	int log_debug = 0;
	int batch = 0;

	int bits = 0;
	int token_slot = 0;
	int selfsign = 0;
	int newkey = 0;

	char * algor_opt = NULL;
	char * hsm_name = NULL;
	char * uri = NULL;
	char * signkey = NULL;
	char * signcert = NULL;
	char * subject = NULL;
	char * profile = NULL;
	char * profileuri = NULL;
	char * oidsuri = NULL;
	char * serial = NULL;
	char * type = NULL;

	int days  = 30;
	int hours = 0;
	int mins  = 0;
	int secs  = 0;

	unsigned long validity = 0;
	int datatype = 0;

	if( argc < 2 ) {
		usage();
	}

	cmd = argv[1];

	for ( i = start; i < argc; i++ ) {
		if ( strncmp_nocase ("-token", argv[i], 5) == 0 ) {
			if( argv[i++] == NULL ) usage();
			token_name = argv[i];
		} else if ( strncmp_nocase("-config", argv[i], 7 ) == 0 ) {
			if( argv[i++] == NULL ) usage();
			config = argv[i];
		} else if ( strncmp_nocase("-type", argv[i], 5 ) == 0 ) {
			if( argv[i++] == NULL ) usage();
			type = argv[i];
		} else if ( strncmp_nocase("-hsm", argv[i], 4 ) == 0 ) {
			if( argv[i++] == NULL ) usage();
			hsm_name = argv[i];
		} else if ( strncmp_nocase("-in", argv[i], 3 ) == 0 ) {
			if( argv[i++] == NULL ) usage();
			infile = argv[i];
		} else if ( strncmp_nocase("-out", argv[i], 4 ) == 0 ) {
			if( argv[i++] == NULL ) usage();
			outfile = argv[i];
		} else if ( strncmp_nocase("-algor", argv[i], 6 ) == 0 ) {
			if( argv[i++] == NULL ) usage();
			algor_opt = argv[i];
		} else if ( strncmp_nocase("-signkey", argv[i], 8 ) == 0 ) {
			if( argv[i++] == NULL ) usage();
			signkey = argv[i];
			uri = signkey;
		} else if ( strncmp_nocase("-signcert", argv[i], 9 ) == 0 ) {
			if( argv[i++] == NULL ) usage();
			signcert = argv[i];
		} else if ( strncmp_nocase("-uri", argv[i], 4 ) == 0 ) {
			if( argv[i++] == NULL ) usage();
			uri = argv[i];
			signkey = uri;
			if( strlen(uri) < 5 ) {
				fprintf(stderr, "\nERROR: URI must be at least "
					"five (5) chars long\n\n");
				return 1;
			}
		} else if ( strncmp_nocase("-subject", argv[i], 8 ) == 0 ) {
			if( argv[i++] == NULL ) usage();
			subject = argv[i];
		} else if ( strncmp_nocase("-serial", argv[i], 7 ) == 0 ) {
			if( argv[i++] == NULL ) usage();
			serial = argv[i];
		} else if ( strncmp_nocase("-profileuri", argv[i], 11 ) == 0 ) {
			if( argv[i++] == NULL ) usage();
			profileuri = argv[i];
		} else if ( strncmp_nocase("-profile", argv[i], 8 ) == 0 ) {
			if( argv[i++] == NULL ) usage();
			profile = argv[i];
		} else if ( strncmp_nocase("-oidsuri", argv[i], 8 ) == 0 ) {
			if( argv[i++] == NULL ) usage();
			oidsuri = argv[i];
		} else if ( strncmp_nocase("-days", argv[i], 5 ) == 0 ) {
			if( argv[i++] == NULL ) usage();
			days = atoi( argv[i] );
		} else if ( strncmp_nocase("-hours", argv[i], 6 ) == 0 ) {
			if( argv[i++] == NULL ) usage();
			hours = atoi( argv[i] );
		} else if ( strncmp_nocase("-mins", argv[i], 5 ) == 0 ) {
			if( argv[i++] == NULL ) usage();
			mins = atoi( argv[i] );
		} else if ( strncmp_nocase("-secs", argv[i], 5 ) == 0 ) {
			if( argv[i++] == NULL ) usage();
			secs = atoi( argv[i] );
		} else if ( strncmp_nocase("-bits", argv[i], 5 ) == 0 ) {
			if( argv[i++] == NULL ) usage();
			bits = atoi(argv[i]);
		} else if ( strncmp_nocase("-slot", argv[i], 5 ) == 0 ) {
			if( argv[i++] == NULL ) usage();
			token_slot = atoi(argv[i]);
		} else if ( strncmp_nocase("-newkey", argv[i], 7 ) == 0 ) {
			newkey = 1;
		} else if ( strncmp_nocase("-batch", argv[i], 6 ) == 0 ) {
			batch = 1;
		} else if ( strncmp_nocase("-selfsign", argv[i], 9 ) == 0 ) {
			selfsign = 1;
		} else if ( strncmp_nocase("-verbose", argv[i], 8 ) == 0 ) {
			verbose = 1;
		} else if ( strncmp_nocase("-debug", argv[i], 6 ) == 0 ) {
			debug = 1;
		} else {
			usage();
		}
	}

	if( verbose ) log_level = PKI_LOG_INFO;
	if( debug ) log_debug |= PKI_LOG_FLAGS_ENABLE_DEBUG;

	if( !outfile || strncmp_nocase(outfile, "stdout", 6 ) == 0 ) {
		outfile = "fd://1";
	} else if ( strncmp_nocase( outfile, "stderr", 6 ) == 0 ) {
		outfile = "fd://2";
	}

	if( !infile || strncmp_nocase(outfile, "stdin", 5 ) == 0 ) {
		infile = "fd://1";
	}

	validity = (unsigned long)(secs + mins * 60 + hours * 3600 + 
					days * 86400);

	if(( PKI_log_init (PKI_LOG_TYPE_STDERR, log_level, NULL,
                        log_debug, NULL )) == PKI_ERR ) {
                exit(1);
        }

	if((tk = PKI_TOKEN_new_null()) == NULL ) {
		printf("ERROR, can not allocate token!\n\n");
		exit(1);
	}

	if( verbose ) {
		printf("Loading Token .. ");
		fflush( stdout );
	}
	if(( PKI_TOKEN_init( tk, config, token_name )) == PKI_ERR) {
		printf("ERROR, can not load token (enable debug for "
							"details)!\n\n");
		exit(1);
	}
	if( verbose ) printf("Ok.\n");

	if( token_name == NULL ) {
		if( hsm_name ) {
			if( verbose ) {
				printf("Loading HSM (%s) .. ", hsm_name );
				fflush( stdout );
			}
			if((tk->hsm = HSM_new ( config, hsm_name )) == NULL ) {
				printf("ERROR, can not load HSM (enable debug"
						" for details)!\n\n");
				exit(1);
			};
			if( verbose ) printf ( "Ok.\n");

			tk->type = tk->hsm->type;
		};

		if ( signkey ) {
			if( verbose ) printf("Loading KeyPair...");
			fflush( stdout );

			if((PKI_TOKEN_load_keypair( tk, uri )) == PKI_ERR){
					printf("\nERROR, can not load key [%s]\n\n", signkey );
					exit(1);
			}
		}

		if ( signcert ) {
			if( verbose ) printf("Loading SignCert...");
			fflush( stdout );

			if((PKI_TOKEN_load_cert( tk, signcert )) == PKI_ERR) {
					fprintf(stderr, "ERROR: Can not load signcert (%s)\n\n",
						signcert );
					exit(1);
			}
		}

	}

	if((PKI_TOKEN_use_slot( tk, token_slot )) == PKI_ERR ) {
		printf("ERROR, can not use selected slot (%d/%s)\n", 
						token_slot, token_name );
		exit (1);
	}

	if( !batch ) {
		printf("\n\nUsing Token (%s::%d)\n", 
					token_name ? token_name : "none", 
						token_slot );
		if( verbose )
			printf("  - Config Dir .. : %s\n\n", 
				PKI_TOKEN_get_config_dir ( tk ) ?
				PKI_TOKEN_get_config_dir ( tk ) : "none" );

	}

	if ( oidsuri ) {
		if ( tk->oids != NULL ) {
			PKI_CONFIG_free ( tk->oids );
		};

		if((tk->oids = PKI_CONFIG_OID_load( oidsuri )) == NULL ) {
        	PKI_log_debug("ERROR, can not load oids uri (%s)!\n", oidsuri);
    	};
	};

	if ( profileuri ) {
		PKI_X509_PROFILE *p = NULL;
		if(( p = PKI_X509_PROFILE_load ( profileuri )) == NULL ) {
			// Can not load the specified profile
			fprintf( stderr, "Can not load profile (%s)\n",
				profileuri );
			exit(1);
		};

		PKI_log_debug("Loaded profile: %s (%s)", profileuri,
			PKI_X509_PROFILE_get_name ( p ));

		if( PKI_TOKEN_add_profile( tk, p ) == PKI_ERR ) {
			fprintf( stderr, "ERROR, can not add profile to the Token!\n");
			exit(1);
		};

		PKI_log_debug("Profile Added to Token: Ok.");
	};

	/*
	if( tk ) {
		PKI_TOKEN_login ( tk );
	};
	*/

	if ( strncmp_nocase("version", cmd, 3) == 0 ) {
		version();
	} else if( strncmp( cmd, "clear", 5) == 0 ) {
		printf("\nThis will delete all the contents of the Token.\n");
		if( !batch ) {
			prompt = prompt_str ("Are you sure [y/N] ? ");
		}
		if( batch || (strncmp_nocase(prompt, "y", 1 ) == 0)) {
			printf("\nClearing up the Token Slot .. ");
			fflush( stdout );
			if((HSM_SLOT_clear ( (unsigned long) tk->slot_id, tk->cred, 
							tk->hsm )) == PKI_OK ) {
				printf("Ok.\n\n");
				exit(0);
			} else {
				printf("Error.\n\n");
				exit(1);
			}
		} else {
			printf("\nOperation Cancelled.\n\n");
		}
	} else if( strncmp( cmd, "delete", 6) == 0 ) {
		if( uri == NULL ) {
			printf("ERROR: please use '-uri' to identify what to "
								"delete.\n\n");
			usage();
		}

		if( type == NULL ) {
			/*
			printf("ERROR: please use '-type' to identify the "
					"type of object to delete.\n\n");
			usage();
			*/
			type = "any";
		}

		if( strncmp_nocase( type, "any", 3 ) == 0 ) {
			datatype = PKI_DATATYPE_ANY;
		} else if ( strncmp_nocase(type, "key", 3 ) == 0 ) {
			datatype = PKI_DATATYPE_X509_KEYPAIR;
		} else if ( strncmp_nocase(type, "pubkey", 6 ) == 0 ) {
			datatype = PKI_DATATYPE_PUBKEY;
		} else if ( strncmp_nocase(type, "privkey", 7 ) == 0 ) {
			datatype = PKI_DATATYPE_PRIVKEY;
		} else if ( strncmp_nocase(type, "other", 5 ) == 0 ) {
			datatype = PKI_DATATYPE_X509_OTHER;
		} else if ( strncmp_nocase(type, "trusted", 7 ) == 0 ) {
			datatype = PKI_DATATYPE_X509_TRUSTED;
		} else if ( strncmp_nocase(type, "cacert", 6 ) == 0 ) {
			datatype = PKI_DATATYPE_X509_CA;
		} else if ( strncmp_nocase(type, "cert", 4 ) == 0 ) {
			datatype = PKI_DATATYPE_X509_CERT;
		} else if ( strncmp_nocase(type, "req", 3 ) == 0 ) {
			datatype = PKI_DATATYPE_X509_REQ;
		} else if ( strncmp_nocase(type, "crl", 3 ) == 0 ) {
			datatype = PKI_DATATYPE_X509_CERT;
		} else {
			printf("ERROR: type is not of a recognized type. "
			    "please use one of any, pubkey, privkey, cert, "
				"req, or crl.\n\n");
			exit(1);
		}
		
		printf("\nThis will delete object %s (%s) in Token %s.\n",
						uri, type, token_name );
		if( !batch ) {
			prompt = prompt_str ("Are you sure [y/N] ? ");
		}
		if( batch || (strncmp_nocase(prompt, "y", 1 ) == 0)) {
			URL *url = NULL;

			if((url = URL_new(uri)) == NULL ) {
				printf("ERROR, %s is not a valid "
						"URI.\n\n", uri );
				exit(1);
			}

			printf("\nDeleting Object .. ");
			fflush( stdout );
			if ( datatype == PKI_DATATYPE_X509_KEYPAIR ) {
				int rv, rv2;

				rv = PKI_TOKEN_del_url(tk, url,
						PKI_DATATYPE_PUBKEY );

				rv2 = PKI_TOKEN_del_url( tk, url,
						PKI_DATATYPE_PRIVKEY );

				if( (rv == PKI_OK) && (rv2 == PKI_OK)) {
					printf("Ok.\n\n");
					exit(0);
				} else {
					printf("Error.\n\n");
					exit(1);
				}
			} else {
				if((PKI_TOKEN_del_url(tk, url, 
						datatype)) == PKI_OK ) {
					printf("Ok.\n\n");
					exit(0);
				} else {
					printf("Error.\n\n");
					exit(1);
				}
			}
		} else {
			printf("\nOperation Cancelled.\n\n");
		}
	} else if ( strncmp_nocase(cmd, "info", 4) == 0 ) {
		printf("\nPrinting Token information (%s):\n\n", 
			token_name ? token_name : "default" );
		PKI_TOKEN_print_info ( tk );

	} else if ( strncmp_nocase(cmd, "list", 4) == 0 ) {
		PKI_TOKEN_STACK *sk = NULL;
		PKI_TOKEN *tk = NULL;

		printf("\nAvailable (configured) Tokens:\n");
		if((sk = PKI_get_all_tokens( config )) == NULL ) {
			printf("- None.\n\n");
			exit(0);
		}

		printf("\nFound %d Tokens:\n", PKI_STACK_elements (sk));

		while( (tk = PKI_STACK_TOKEN_pop ( sk )) != NULL ) {
			printf("- %s (in %s)\n", PKI_TOKEN_get_name ( tk ), 
				PKI_TOKEN_get_config_dir ( tk ) );
			if( tk ) PKI_TOKEN_free ( tk );
		}
		printf("\n");

		if( sk ) PKI_STACK_TOKEN_free ( sk );

	} else if ( strncmp_nocase(cmd, "genkey", 6) == 0 ) {
		if((gen_keypair ( tk, bits, uri, algor_opt, batch )) 
								== PKI_ERR ) {
			printf("\nERROR, can not create keypair!\n\n");
			exit(1);
		}
	} else if ( strncmp_nocase(cmd, "genreq", 6) == 0 ) {
		/* We need to generate a new keypair first - if the '-newkey'
		   switch is used */
		if( newkey ) {
			if( verbose ) printf("Generating KeyPair...");
			if((gen_keypair (tk, bits, uri, algor_opt, batch)) 
								== PKI_ERR ) {
				printf("\nERROR, can not create keypair!\n\n");
				exit(1);
			}
			if( verbose ) printf("Ok.\n");
		}

		if( verbose ) printf("Generating new request (%s) ... ", 
				subject ? subject : "no subject" );
		fflush(stdout);
		if((PKI_TOKEN_new_req( tk, subject, NULL )) == PKI_ERR) {
			printf("ERROR, can not generate a new Request!\n");
			exit(1);
		};
		if( verbose ) printf("Ok.\n");

		if( outfile ) {
			if (verbose ) {
				printf("Writing request to [%s]\n", outfile);
				fflush(stdout);
			}

			if((PKI_TOKEN_export_req( tk, outfile,
						PKI_DATA_FORMAT_PEM )) == PKI_ERR ) {
				printf("ERROR, can not save req!\n");
				exit(1);
			}
			if( verbose ) {
				printf("Ok.\n");
			}
		}

	} else if ( strncmp_nocase(cmd, "gencert", 7) == 0 ) {

		if( signkey ) {
			if(verbose) {
				printf("Loading Keypair (%s) ... ", signkey);
				fflush(stdout);
			}
			if((PKI_TOKEN_load_keypair( tk, signkey )) 
							== PKI_ERR){
				printf("\nERROR, can not load key "
						"[%s]\n\n", signkey );
				exit(1);
			} else {
				if( !tk->keypair ) {
					printf("\nERROR, can not load "
						"keypair from token "
						"config!\n\n");
					exit(1);
				}
				if( verbose ) printf("Ok.\n");
			}
		} else {
			if ( tk->keypair == NULL ) {
				printf("\nERROR, no keypair loaded - check "
					"token config!\n\n");
				exit(1);
			}
		}

		if ( signcert ) {
			if(verbose) {
				printf("Loading Signing Cert (%s) ... ", signcert);
				fflush(stdout);
			};
			if((PKI_TOKEN_load_cert( tk, signcert )) == PKI_ERR){
				printf("\nERROR, can not load signing cert "
						"[%s]\n\n", signcert );
				exit(1);
			};
		} else {
			if ( (tk->cert == NULL) && (!selfsign) ) {
				fprintf(stderr, "ERROR, no signing cert provided!\n");
				exit ( 1 );
			}
		}
		if( verbose ) printf("* Generating a new Certificate:\n");

		if ( !infile ) {
			printf("\nERROR, '-in <req>' is required!\n\n");
			exit(1);
		}

		/*
		printf("\nThis will delete all the contents of the Token.\n");
		if( !batch ) {
			prompt = prompt_str ("Are you sure [y/N] ? ");
		}
		*/

		if( PKI_TOKEN_load_req ( tk, infile ) == PKI_ERR ) {
			printf("\nERROR, can not load request %s!\n\n", infile);
			return ( 1 );
		}

		if ( algor_opt ) {
			int algor_id;
			PKI_ALGOR *algor = NULL;

			if((algor = PKI_ALGOR_get_by_name ( algor_opt )) == NULL ) {
				PKI_log_err ("Can not set algor to %s", algor_opt);
				return(1);
			}

			algor_id = PKI_ALGOR_get_id (algor);
			if( PKI_TOKEN_set_algor ( tk, algor_id ) == PKI_ERR ) {
				PKI_log_err( "Can not set algor in Token (%d)", algor_id);
				return(1);
			}
		};

		if ( selfsign == 1 ) {
			if ( verbose ) printf ("  - Self Signing "
							"certificate .... ");
			if((PKI_TOKEN_self_sign( tk, subject, serial, 
					validity, profile )) == PKI_ERR ) {
				printf("ERROR, can not self sign certificate!\n");
				return(0);
			}
			if ( verbose ) printf("Ok.\n");

			if(verbose) {
				printf("  - Writing Certificate to (%s)... ",
								outfile );
				fflush(stdout);
			};

			if((PKI_TOKEN_export_cert( tk, outfile,
					PKI_DATA_FORMAT_PEM )) == PKI_ERR ) {
				printf("ERROR,can not save cert in "
							"certificate.pem!\n");
				return(0);
			}
			if ( verbose ) printf("Ok.\n");
		} else {
			PKI_X509_CERT * x = NULL;

			if(verbose ) printf("  - Issuing new certificate ... ");
			fflush(stdout);

			if( tk->cert == NULL ) {
				printf("ERROR, the token has no usable "
						"(signing) certificate!\n");
				if( tk ) PKI_TOKEN_free (tk);
				return ( 1 );
			}

			if((x = PKI_TOKEN_issue_cert( tk, subject, serial, 
					validity, tk->req, profile)) == NULL ) {
				printf("ERROR, can not issue certificate!\n\n");
				fflush(stdout);
				if ( tk ) PKI_TOKEN_free ( tk );
				return (1);
			}
			if(verbose) printf("Ok.\n");

			if(verbose) {
				printf("  - Writing Certificate to (%s)... ",
								outfile );
				fflush(stdout);
			};

			if((PKI_X509_CERT_put ( x, PKI_DATA_FORMAT_PEM, 
				outfile, NULL, tk->cred, tk->hsm)) == PKI_ERR) {
				printf("ERROR, can not save cert in "
						"%s\n", outfile );
				return(0);
			}
			if ( verbose ) printf("Ok.\n");
		}
	} else if ( strncmp_nocase(cmd, "gencrl", 6) == 0 ) {
		PKI_X509_CRL *crl = NULL;

		if( !tk->keypair ) {
			fprintf(stderr, "ERROR: no SignKey loaded\n\n");
			exit(1);
		}

		if( !tk->cert ) {
			fprintf(stderr, "ERROR: no SignCert loaded\n\n");
			exit(1);
		}

		if ( !outfile ) {
			outfile = "fd://1";
		}

		if( verbose ) printf("Generating new CRL (%s) ... ", outfile );
		fflush(stdout);

		if((crl = PKI_TOKEN_issue_crl( tk, serial, validity, NULL,
				profile )) == NULL) {
			fprintf(stderr, "ERROR: can not generate a new CRL!\n\n");
			exit(1);
		};
		if((crl = PKI_X509_CRL_put ( crl, PKI_DATA_FORMAT_PEM, 
					outfile, NULL, NULL )) == NULL ) {
			fprintf(stderr, "ERROR: can not save crl!\n\n");
			exit(1);
		}
		if( verbose ) printf("Ok.\n");

	} else if ( strncmp_nocase(cmd, "delete", 6) == 0 ) {
		if( uri == NULL ) {
			printf("\nERROR, no '-uri' provided!");
			usage();
		}
	} else if ( strncmp_nocase(cmd, "import", 6) == 0 ) {
		if ( type == NULL ) {
			printf("\nERROR, no <type> provided!");
			usage();
		}

		if( !uri ) {
			printf("ERROR, no '-uri' provided!\n\n");
			return ( 1 );
		}

		if(strncmp_nocase(type, "user", 4) == 0 ) {
			PKI_X509_CERT_STACK * sk = NULL;

			if(verbose) {
				printf("Importing User certificate(s) %s ... ",
						infile );
				fflush ( stdout );
			}


			if((sk = PKI_X509_CERT_STACK_get ( infile, 
						tk->cred, NULL )) == NULL ) {
				printf("ERROR, can not load cert (%s)!\n\n",
						infile );
				return ( 1 );
			}

			if(PKI_TOKEN_import_cert_stack(tk, sk, 
					PKI_DATATYPE_X509_CERT, uri) 
								== PKI_ERR) {
				printf("ERROR!\n\n");
				exit(1);
			};

			if( verbose ) {
				printf("Ok.\n");
			}
		} else if(strncmp_nocase(type, "key", 3) == 0 ) {

			PKI_X509_KEYPAIR *x = NULL;

			if(verbose) {
				printf("Importing key %s ... ",
						infile );
				fflush ( stdout );
			}

			if((x = PKI_X509_KEYPAIR_get ( infile, 
						tk->cred, NULL)) == NULL ) {
				printf("ERROR, can not load keyfile (%s)\n\n",
					infile);
				return ( 1 );
			}

			if(PKI_TOKEN_import_keypair(tk, x, uri) ==PKI_ERR) {
				printf("ERROR!\n\n");
				exit(1);
			};

			printf("Not implemented Yet!\n");

			if( verbose ) {
				printf("Ok.\n");
			}
		} else if(strncmp_nocase(type, "ca", 2) == 0 ) {
			PKI_X509_CERT_STACK *sk = NULL;

			if(verbose) {
				printf("Importing CA certificate(s) %s ... ",
						infile );
				fflush ( stdout );
			}


			if((sk = PKI_X509_CERT_STACK_get ( infile, 
						tk->cred, NULL )) == NULL ) {
				printf("ERROR, can not load cert (%s)!\n\n",
						infile );
				return ( 1 );
			}

			if(PKI_TOKEN_import_cert_stack ( tk, sk, 
					PKI_DATATYPE_X509_CA, uri) 
								== PKI_ERR) {
				printf("ERROR!\n\n");
				exit(1);
			};

			if( verbose ) {
				printf("Ok.\n");
			}
		} else if(strncmp_nocase(type, "trusted", 7) == 0 ) {
			PKI_X509_CERT_STACK *sk = NULL;

			if(verbose) {
				printf("Importing Trusted (TA) certificate(s) "
						"%s ... ", infile );
				fflush ( stdout );
			}


			if((sk = PKI_X509_CERT_STACK_get ( infile, tk->cred, 
							NULL )) == NULL ) {
				printf("ERROR, can not load cert (%s)!\n\n",
						infile );
				return ( 1 );
			}

			if(PKI_TOKEN_import_cert_stack(tk, sk, 
					PKI_DATATYPE_X509_TRUSTED, uri) 
								== PKI_ERR) {
				printf("ERROR!\n\n");
				exit(1);
			};

			if( verbose ) {
				printf("Ok.\n");
			}
		} else if(strncmp_nocase(type, "other", 5) == 0 ) {
			PKI_X509_CERT_STACK *sk = NULL;

			if(verbose) {
				printf("Importing Other certificate(s) "
						"%s ... ", infile );
				fflush ( stdout );
			}


			if((sk = PKI_X509_CERT_STACK_get ( infile, tk->cred, 
							NULL )) == NULL ) {
				printf("ERROR, can not load cert (%s)!\n\n",
						infile );
				return ( 1 );
			}

			if(PKI_TOKEN_import_cert_stack ( tk, sk, 
					PKI_DATATYPE_X509_OTHER, uri) 
								== PKI_ERR) {
				printf("ERROR!\n\n");
				exit(1);
			};

			if( verbose ) {
				printf("Ok.\n");
			}
		} else {
			printf("\nERROR: type '%s' not recognized!\n\n", type);
			usage();
		}
	} else {
		printf("\n  ERROR: command not recognized (%s)\n", cmd );
		usage();
	}

	if( verbose ) printf("Freeing Token Object ... ");
	fflush(stdout);

	if( tk ) PKI_TOKEN_free ( tk );
	if( prof ) PKI_X509_PROFILE_free ( prof );

	if(verbose) printf("Done.\n\n");

	return (0);

	/*
	printf("Setting Token Algorithm to DSA-SHA1 ... ");
	if((PKI_TOKEN_set_algor ( tk, PKI_ALGOR_DSA_SHA1 )) == PKI_ERR ) {
		printf("ERROR, can not set the DSA-SHA1 crypto scheme!\n");
	}
	printf("Ok.\n");
	*/

	/*
	printf("Setting Token Algorithm to DSA-SHA256 ... ");
	if((PKI_TOKEN_set_algor ( tk, PKI_ALGOR_DSA_SHA256 )) == PKI_ERR ) {
		printf("ERROR, can not set the RSA crypto scheme!\n");
	}
	printf("Ok.\n");

	printf("Setting Token Algorithm to RSA-MD5 ... ");
	if((PKI_TOKEN_set_algor ( tk, PKI_ALGOR_RSA_MD5 )) == PKI_ERR ) {
		printf("ERROR, can not set the RSA crypto scheme!\n");
	}
	printf("Ok.\n");
	*/

	/*
	if((PKI_TOKEN_new_req( tk, "CN=Test4, O=OpenCA", "test" )) == PKI_ERR) {
		printf("ERROR, can not generate a new Request!\n");
		return(0);
	}
	printf("Ok.\n");

	printf("* Writing request to request.pem .... ");
	if((PKI_TOKEN_export_req( tk, "request.pem",
			PKI_DATA_FORMAT_PEM )) == PKI_ERR ) {
		printf("ERROR, can not save req pkcs11_request.pem!\n");
		return(0);
	}
	printf("Ok.\n");
	*/


	/*
	if((PKI_TOKEN_set_scheme ( tk, PKI_SCHEME_RSA )) == PKI_ERR ) {
		printf("ERROR, can not set the RSA crypto scheme!\n");
		return (0);
	}

	if((PKI_TOKEN_set_algor ( tk, PKI_ALGOR_RSA_SHA1 )) == PKI_ERR ) {
		printf("ERROR, can not set the RSA crypto scheme!\n");
		return (0);
	}

	if((PKI_TOKEN_new_keypair ( tk, 1024, NULL )) == PKI_ERR) {
		printf("ERROR, can not generate new keypair!\n");
		return (0);
	}

	printf("* Generating new Request ... ");
	if((PKI_TOKEN_new_req( tk, "CN=Test4, O=OpenCA", "test" )) == PKI_ERR) {
		printf("ERROR, can not generate a new Request!\n");
		return(0);
	}
	printf("Ok.\n");

	printf("* Writing request to results/test4_req1.pem .... ");
	if((PKI_TOKEN_write_req( tk, "results/test4_req1.pem",
			PKI_DATA_FORMAT_PEM )) == PKI_ERR ) {
		printf("ERROR, can not save req results/test4_req1.pem!\n");
		return(0);
	}
	printf("Ok.\n");

	printf("* Self Signing certificate .... ");
	if((PKI_TOKEN_self_sign( tk, NULL, NULL, "Test" )) == PKI_ERR ) {
		printf("ERROR, can not self sign certificate!\n");
		return(0);
	}
	printf("Ok.\n");

	if((PKI_KEYPAIR_export( tk->keypair, NULL, PKI_DATA_FORMAT_PEM, NULL, 
			keyfile,  NULL )) == PKI_ERR ) {
                printf("ERROR::Can not export key (%s)!\n", buf );
                return(0);
        };

	printf("Writing Certificate to file... \n");
	if((PKI_TOKEN_write_cert( tk, "results/test4_cert1.pem",
			PKI_DATA_FORMAT_PEM )) == PKI_ERR ) {
		printf("ERROR,can not save cert in results/test4_cert1.pem!\n");
		return(0);
	}
	*/

}

